function belief_propagation

% Carlo Dal Mutto
%
% dalmutto@dei.unipd.it


clc
clear
close all
tic

%% GLOBAL VARIABLES
global SCENE
global MIN_ROW_D
global MAX_ROW_D
global MIN_COL_D
global MAX_COL_D
global DEBUG % if DEBUG==true -> the messages received by the pixel (DEBUG_MESSAGES_ROW,DEBUG_MESSAGES_COL)
MIN_ROW_D = 1;
MAX_ROW_D = 778; 
MIN_COL_D = 1;
MAX_COL_D = 1032;
SCENE = 1;
DEBUG = false;
SUPER_RES = 1;


%% ALGORITHM PARAMETERS

NO_FRAME = true; % if true, the cost volume is cropped, leaving the first 2 and last 2 rows and the first 2 and last 2 cols (in order not to account for them if they are 0)
BELIEF_PROPAGATION_LOOPS = 10; % # of loops of Loopy Belief Propagation
BELIEF_PROPAGATION_TH = 0.1;
BELIEF_PROPAGATION_TEMPERATURE = sqrt(0.0001); % 0.001
TH_BOOST = 1;
KERNEL_EXP = 2;

if DEBUG
    DEBUG_MESSAGES_ROW = MIN_ROW_D-1+11; %#ok<UNRCH>
    DEBUG_MESSAGES_COL = MIN_COL_D-1+9;
end
if NO_FRAME 
    OFFSET = 2*SUPER_RES; % Offset that is used in order to properly load the data
    N_ROWS = MAX_ROW_D - 2*OFFSET;
    N_COLS = MAX_COL_D - 2*OFFSET;
else
    N_ROWS = MAX_ROW_D - 2*OFFSET;
    N_COLS = MAX_COL_D - 2*OFFSET;
    OFFSET = 0;
end

%%  LOADING FUSION CONFIGURATION
load dataTerms
CALIB_PARS_FOLDER = 'C:\Users\Carlo\Desktop\Other_Methods\Data\Calibration_Parameters_2\';
K_S = load([CALIB_PARS_FOLDER,'intrinseci_Rett_dx.txt']); % Rectified stereo cameras intrinsics
focal = K_S(1,1);
Rt_S2T = load([CALIB_PARS_FOLDER,'Rt_Rett_sx_dx.txt']); % This is loaded only for calculating the baseline
baseline = abs(Rt_S2T(1,4));
bf = baseline*focal;
depthTable = bf./disparityTable;

%%
interval_length = zeros(N_ROWS,N_COLS);
tof_length = zeros(N_ROWS,N_COLS);
stereo_length = zeros(N_ROWS,N_COLS);

% Length of the ToF and the stereo intervals
for i_r = 1:N_ROWS
    for i_c = 1:N_COLS
        interval_length(i_r,i_c) = length(disparityTable);
        tof_length(i_r,i_c) = length(disparityTable);
        stereo_length(i_r,i_c) = length(disparityTable);
    end
end

disp(['Average number of samples: ',num2str(mean(interval_length(:)))])

%% Calculating probability cell
probability = cell(N_ROWS,N_COLS);
interval_cell = cell(N_ROWS,N_COLS);
tof_probability = cell(N_ROWS,N_COLS);
stereo_probability = cell(N_ROWS,N_COLS);
for i_r = 1:N_ROWS
    for i_c = 1:N_COLS
        probability{i_r,i_c} = zeros(1,interval_length(i_r,i_c));
        tof_probability{i_r,i_c} = zeros(1,interval_length(i_r,i_c));
        stereo_probability{i_r,i_c} = zeros(1,interval_length(i_r,i_c));
        interval_cell{i_r,i_c} = zeros(1,interval_length(i_r,i_c));
    end
end

for i_r = 1:N_ROWS
    for i_c = 1:N_COLS
        tof_probability{i_r,i_c}(:) = costVolT(i_r,i_c,:);
        stereo_probability{i_r,i_c}(:) = costVolS(i_r,i_c,:);
        interval_cell{i_r,i_c}(:) = depthTable(:);
        probability{i_r,i_c}(:) = tof_probability{i_r,i_c}(:)...
                                  .*stereo_probability{i_r,i_c}(:)...
                                  / sum(tof_probability{i_r,i_c}(:)...
                                        .*stereo_probability{i_r,i_c}(:));
        stereo_probability{i_r,i_c}(:) = stereo_probability{i_r,i_c}(:)/sum(stereo_probability{i_r,i_c}(:));
        tof_probability{i_r,i_c}(:) = tof_probability{i_r,i_c}(:)/(sum(tof_probability{i_r,i_c}(:)));
    end
end

interval = interval_cell;

%% Calculating the neighborhood
neighborhood = cell(N_ROWS,N_COLS); % Neighborhood struct "neighborhood{i_r,i_c}(neighbor).u"
% 4 corners, only 2 neighbors
neighborhood{1,1} = [struct('v',1,'u',2), struct('v',2,'u',1)];
neighborhood{1,N_COLS} = [struct('v',1,'u',N_COLS-1), struct('v',2,'u',N_COLS)];
neighborhood{N_ROWS,1} = [struct('v',N_ROWS-1,'u',1), struct('v',N_ROWS,'u',2)];
neighborhood{N_ROWS,N_COLS} = [struct('v',N_ROWS-1,'u',N_COLS), struct('v',N_ROWS,'u',N_COLS-1)];
% first and last rows
for i_c = 2:N_COLS-1
    neighborhood{1,i_c} = [struct('v',1,'u',i_c-1), struct('v',1,'u',i_c+1), struct('v',2,'u',i_c)]; % first row
    neighborhood{N_ROWS,i_c} = [struct('v',N_ROWS-1,'u',i_c), struct('v',N_ROWS,'u',i_c-1), struct('v',N_ROWS,'u',i_c+1)]; % last row
end
% first and last columns
for i_r = 2:N_ROWS-1
    neighborhood{i_r,1} = [struct('v',i_r-1,'u',1), struct('v',i_r,'u',2), struct('v',i_r+1,'u',1)]; % first column
    neighborhood{i_r,N_COLS} = [struct('v',i_r-1,'u',N_COLS), struct('v',i_r,'u',N_COLS-1), struct('v',i_r+1,'u',N_COLS)]; % last column
end
% all the other elements (box)
for i_r = 2:N_ROWS-1
    for i_c = 2:N_COLS-1
        neighborhood{i_r,i_c} = [struct('v',i_r-1,'u',i_c), struct('v',i_r,'u',i_c-1), struct('v',i_r,'u',i_c+1), struct('v',i_r+1,'u',i_c)];
    end
end

neighbors_number = ones(N_ROWS,N_COLS)*4; % Number of neighboring points
neighbors_number(1,:) = 3;
neighbors_number(N_ROWS,:) = 3;
neighbors_number(:,1) = 3;
neighbors_number(:,N_COLS) = 3;
neighbors_number(1,1) = 2;
neighbors_number(1,N_COLS) = 2;
neighbors_number(N_ROWS,1) = 2;
neighbors_number(N_ROWS,N_COLS) = 2;

%% 1) Messages matrix up (matrix of the messages that single points receives from the points above)
messages_up = cell(N_ROWS,N_COLS);
for i_r = 1:N_ROWS
    for i_c = 1:N_COLS
        messages_up{i_r,i_c} = ones(1, interval_length(i_r,i_c));
    end
end
% 2) Messages matrix down (matrix of the messages that single points receives from the points below)
messages_down = messages_up;
% 3) Messages matrix left (matrix of the messages that single points receive from the points on the left)
messages_left = messages_up;
% 4) Messages matrix right (matrix of the messages that single points receive from the points on the right)
messages_right = messages_up;
% first row does not receive messages from points above (the message is deleted) -> the
% relative cell is empty
for i_c = 1:N_COLS
     messages_up{1,i_c} = [];
end
% last row does not receive messages from below
for i_c = 1:N_COLS
    messages_down{N_ROWS,i_c} = [];
end
% first column does not receive messages from left
for i_r = 1:N_ROWS
    messages_left{i_r,1} = [];
end
% last column does not receive messages from right
for i_r = 1:N_ROWS
    messages_right{i_r,N_COLS} = [];
end

messages_up_new = messages_up; % updated messages
messages_down_new = messages_down;
messages_left_new = messages_left;
messages_right_new = messages_right;

%% Calculating the current depthmap before the first execution of BP
current_depthmap = zeros(N_ROWS,N_COLS);
for i_r = 1:N_ROWS
    for i_c = 1:N_COLS
        prob_temp = probability{i_r,i_c}(:);
        prob_temp = prob_temp(:);
        [~,idx] = max(prob_temp);
        current_depthmap(i_r,i_c) = interval{i_r,i_c}(idx);
    end
end
% figure()
% imagesc(current_depthmap)
% title('Depthmap before the first iteration')
eval(['save iteration_',num2str(0),' current_depthmap'])
[error,pixel_number] = calculate_error_gt(current_depthmap);
[prob_value, data_prob_value, smooth_prob_value] = belief_propagation_probability_calculation(current_depthmap, interval, probability, neighborhood, BELIEF_PROPAGATION_TH, BELIEF_PROPAGATION_TEMPERATURE);
disp(['Fusion error: ',num2str(error)])
disp(['Pixel number: ',num2str(pixel_number)])
disp(['Belief propagation probability: ',num2str(prob_value)])
disp(['Data probability: ',num2str(data_prob_value)])
disp(['Smoothness probability: ',num2str(smooth_prob_value)])


%% Messages updating 

probability_new = probability; 
error_history = zeros(1,BELIEF_PROPAGATION_LOOPS+1);
prob_history = zeros(1,BELIEF_PROPAGATION_LOOPS+1);
data_prob_history = zeros(1,BELIEF_PROPAGATION_LOOPS+1);
smooth_prob_history = zeros(1,BELIEF_PROPAGATION_LOOPS+1);
error_history(1) = error;
prob_history(1) = prob_value;
data_prob_history(1) = data_prob_value;
smooth_prob_history(1) = smooth_prob_value;

for t = 1:BELIEF_PROPAGATION_LOOPS
    
    % Calculating new messages
    for i_r =MIN_ROW_D:MAX_ROW_D%1:N_ROWS
        clc
        disp(['Belief propagation row ',num2str(i_r),' of ',num2str(N_ROWS),' loop ',num2str(t),' of ',num2str(BELIEF_PROPAGATION_LOOPS)])
        pause(0.01)
        for i_c = MIN_COL_D:MAX_COL_D%1:N_COLS
            
            
            if ~isempty(messages_up{i_r,i_c}) % Messages received from above
                
                interval_temp = interval{i_r,i_c}(:); % Interval of the considered point at this iteration
                interval_temp = interval_temp(:);
                
                interval_temp_aux = interval{i_r-1,i_c}(:); % Interval of the current neighboring points of the considered point at this iteration
                interval_temp_aux = interval_temp_aux(:);
                    
                probability_temp_aux = probability{i_r-1,i_c}(:);
                probability_temp_aux = probability_temp_aux(:);
                
                for i_k = 1:length(interval_temp)
                    message_temp = 0;
                    
                    for i_l = 1:length(interval_temp_aux)
                        message_temp_aux = 1;
                        for i_n = 1:neighbors_number(i_r-1,i_c)
                            u_temp = neighborhood{i_r-1,i_c}(i_n).u;
                            v_temp = neighborhood{i_r-1,i_c}(i_n).v;
                            if (u_temp~=i_c && v_temp~=i_r)
                                if(v_temp<i_r-1)
                                    message_temp_aux = message_temp_aux*messages_up{i_r-1,i_c}(i_l);
                                elseif(v_temp>i_r-1)
                                    message_temp_aux = message_temp_aux*messages_down{i_r-1,i_c}(i_l);
                                elseif(u_temp<i_c) 
                                    message_temp_aux = message_temp_aux*messages_left{i_r-1,i_c}(i_l);
                                elseif(u_temp>i_c)
                                    message_temp_aux = message_temp_aux*messages_right{i_r-1,i_c}(i_l);
                                end
                            end
                        end
                        if abs(interval_temp(i_k)-interval_temp_aux(i_l))<BELIEF_PROPAGATION_TH
                            
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                                          exp(-abs(interval_temp(i_k)-interval_temp_aux(i_l))^KERNEL_EXP/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                                          message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        else
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                                          exp(-BELIEF_PROPAGATION_TH^KERNEL_EXP*TH_BOOST/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                                          message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        end
                    end
                    messages_up_new{i_r,i_c}(i_k) = message_temp; 
                end
            end    
            
            
            if ~isempty(messages_down{i_r,i_c}) % Messages received from below
                
                interval_temp = interval{i_r,i_c}(:); % Interval of the considered point at this iteration
                interval_temp = interval_temp(:);
                
                interval_temp_aux = interval{i_r+1,i_c}(:); % Interval of the current neighboring points of the considered point at this iteration
                interval_temp_aux = interval_temp_aux(:);
                
                probability_temp_aux = probability{i_r+1,i_c}(:);
                probability_temp_aux = probability_temp_aux(:);
                    

                for i_k = 1:length(interval_temp)  
                    message_temp = 0;
                    for i_l = 1:length(interval_temp_aux)
                        
                        message_temp_aux = 1;
                        for i_n = 1:neighbors_number(i_r+1,i_c)
                            u_temp = neighborhood{i_r+1,i_c}(i_n).u;
                            v_temp = neighborhood{i_r+1,i_c}(i_n).v;
                            if (u_temp~=i_c && v_temp~=i_r)
                                if(v_temp<i_r+1)
                                    message_temp_aux = message_temp_aux*messages_up{i_r+1,i_c}(i_l);
                                elseif(v_temp>i_r+1)
                                    message_temp_aux = message_temp_aux*messages_down{i_r+1,i_c}(i_l);
                                elseif(u_temp<i_c) 
                                    message_temp_aux = message_temp_aux*messages_left{i_r+1,i_c}(i_l);
                                elseif(u_temp>i_c)
                                    message_temp_aux = message_temp_aux*messages_right{i_r+1,i_c}(i_l);
                                end
                            end
                        end
                        if abs(interval_temp(i_k)-interval_temp_aux(i_l))<BELIEF_PROPAGATION_TH

                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                exp(-abs(interval_temp(i_k)-interval_temp_aux(i_l))^KERNEL_EXP/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        else
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                exp(-BELIEF_PROPAGATION_TH^KERNEL_EXP*TH_BOOST/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        end

                    end
                    messages_down_new{i_r,i_c}(i_k) = message_temp; 
                end
                
            end
            
            if ~isempty(messages_left{i_r,i_c}) % Messages received from left
               
                interval_temp = interval{i_r,i_c}(:); % Interval of the considered point at this iteration
                interval_temp = interval_temp(:);
                
                interval_temp_aux = interval{i_r,i_c-1}(:); % Interval of the current neighboring points of the considered point at this iteration
                interval_temp_aux = interval_temp_aux(:);
                
                probability_temp_aux = probability{i_r,i_c-1}(:);
                probability_temp_aux = probability_temp_aux(:);
                
                for i_k = 1:length(interval_temp)
                    
                    message_temp = 0;
                    
                    for i_l = 1:length(interval_temp_aux)
                        message_temp_aux = 1;
                        for i_n = 1:neighbors_number(i_r,i_c-1)
                            u_temp = neighborhood{i_r,i_c-1}(i_n).u;
                            v_temp = neighborhood{i_r,i_c-1}(i_n).v;
                            if (u_temp~=i_c && v_temp~=i_r)
                                if(v_temp<i_r)
                                    message_temp_aux = message_temp_aux*messages_up{i_r,i_c-1}(i_l);
                                elseif(v_temp>i_r)
                                    message_temp_aux = message_temp_aux*messages_down{i_r,i_c-1}(i_l);
                                elseif(u_temp<i_c-1) 
                                    message_temp_aux = message_temp_aux*messages_left{i_r,i_c-1}(i_l);
                                elseif(u_temp>i_c-1)
                                    message_temp_aux = message_temp_aux*messages_right{i_r,i_c-1}(i_l);
                                end
                            end
                        end

                        if abs(interval_temp(i_k)-interval_temp_aux(i_l))<BELIEF_PROPAGATION_TH
                            
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                                          exp(-abs(interval_temp(i_k)-interval_temp_aux(i_l))^KERNEL_EXP/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                                          message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        else
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                                          exp(-BELIEF_PROPAGATION_TH^KERNEL_EXP*TH_BOOST/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                                          message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        end

                    end
                    messages_left_new{i_r,i_c}(i_k) = message_temp; 
                end
            end
            
            if ~isempty(messages_right{i_r,i_c}) % Messages received from right
               
                
                interval_temp = interval{i_r,i_c}(:); % Interval of the considered point at this iteration
                interval_temp = interval_temp(:);
                
                interval_temp_aux = interval{i_r,i_c+1}(:); % Interval of the current neighboring points of the considered point at this iteration
                interval_temp_aux = interval_temp_aux(:);
                
                probability_temp_aux = probability{i_r,i_c+1}(:);
                probability_temp_aux = probability_temp_aux(:);
                       
                for i_k = 1:length(interval_temp)
                    
                    message_temp = 0;
                    
                    for i_l = 1:length(interval_temp_aux)
                        message_temp_aux = 1;
                        for i_n = 1:neighbors_number(i_r,i_c+1)
                            u_temp = neighborhood{i_r,i_c+1}(i_n).u;
                            v_temp = neighborhood{i_r,i_c+1}(i_n).v;
                            if (u_temp~=i_c && v_temp~=i_r)
                                if(v_temp<i_r)
                                    message_temp_aux = message_temp_aux*messages_up{i_r,i_c+1}(i_l);
                                elseif(v_temp>i_r)
                                    message_temp_aux = message_temp_aux*messages_down{i_r,i_c+1}(i_l);
                                elseif(u_temp<i_c+1) 
                                    message_temp_aux = message_temp_aux*messages_left{i_r,i_c+1}(i_l);
                                elseif(u_temp>i_c+1)
                                    message_temp_aux = message_temp_aux*messages_right{i_r,i_c+1}(i_l);
                                end
                            end
                        end

                        if abs(interval_temp(i_k)-interval_temp_aux(i_l))<BELIEF_PROPAGATION_TH
                            
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                                          exp(-abs(interval_temp(i_k)-interval_temp_aux(i_l))^KERNEL_EXP/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                                          message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        else
                            message_temp = message_temp + probability_temp_aux(i_l)*... % Data term
                                                          exp(-BELIEF_PROPAGATION_TH^KERNEL_EXP*TH_BOOST/BELIEF_PROPAGATION_TEMPERATURE^KERNEL_EXP) *... % Smoothness term
                                                          message_temp_aux; % Old messages from the neighbors of the point except for the considered point
                        end

                    end
                    messages_right_new{i_r,i_c}(i_k) = message_temp; 
                end
            end
            
            % alpha_ij
            messages_up_new{i_r,i_c}(:) = messages_up_new{i_r,i_c}(:)./max(messages_up_new{i_r,i_c}(:));
            messages_down_new{i_r,i_c}(:) = messages_down_new{i_r,i_c}(:)./max(messages_down_new{i_r,i_c}(:));
            messages_left_new{i_r,i_c}(:) = messages_left_new{i_r,i_c}(:)./max(messages_left_new{i_r,i_c}(:));
            messages_right_new{i_r,i_c}(:) = messages_right_new{i_r,i_c}(:)./max(messages_right_new{i_r,i_c}(:));
                     
        end
    end
    
    % Calculating new probability
    disp('Calculating new probability')
    for i_r = MIN_ROW_D:MAX_ROW_D%1:N_ROWS
        for i_c = MIN_COL_D:MAX_COL_D%1:N_COLS
            for i_k = 1:length(probability_new{i_r,i_c})
                message_temp = 1;
                if ~isempty(messages_up_new{i_r,i_c})
                    message_temp = message_temp*messages_up_new{i_r,i_c}(i_k);
                end
                if ~isempty(messages_down_new{i_r,i_c})
                    message_temp = message_temp*messages_down_new{i_r,i_c}(i_k);
                end
                if ~isempty(messages_left_new{i_r,i_c})
                    message_temp = message_temp*messages_left_new{i_r,i_c}(i_k);
                end
                if ~isempty(messages_right_new{i_r,i_c})
                    message_temp = message_temp*messages_right_new{i_r,i_c}(i_k);
                end
                probability_new{i_r,i_c}(i_k) = probability{i_r,i_c}(i_k) * message_temp;
            end
        end
    end
                
    for i_r = MIN_ROW_D:MAX_ROW_D%1:N_ROWS
        for i_c = MIN_COL_D:MAX_COL_D%1:N_COLS
            prob_temp = probability_new{i_r,i_c}(:);
            prob_temp = prob_temp(:);
            [~,idx] = max(prob_temp);
            current_depthmap(i_r,i_c) = interval{i_r,i_c}(idx);
        end
    end

    [error,pixel_number] = calculate_error_gt(current_depthmap);
    [prob_value, data_prob_value, smooth_prob_value] = belief_propagation_probability_calculation(current_depthmap, interval, probability, neighborhood, BELIEF_PROPAGATION_TH, BELIEF_PROPAGATION_TEMPERATURE);
    disp(['Fusion error: ',num2str(error)])
    disp(['Pixel number: ',num2str(pixel_number)])
    disp(['Belief propagation probability: ',num2str(prob_value)])
    disp(['Data probability: ',num2str(data_prob_value)])
    disp(['Smoothness probability: ',num2str(smooth_prob_value)])
    error_history(t+1) = error;
    prob_history(t+1) = prob_value;
    data_prob_history(t+1) = data_prob_value;
    smooth_prob_history(t+1) = smooth_prob_value;

    eval(['save iteration_',num2str(t),' current_depthmap error_history'])

    % Updating current messages
    messages_up = messages_up_new;
    messages_down = messages_down_new;
    messages_left = messages_left_new;
    messages_right = messages_right_new;
    
    figure(100)
    subplot(4,1,1)
    stem(error_history)
    title('Error before the begin of any LBP iteration')
    xlabel('Iteration')
    ylabel('Error')
    subplot(4,1,2)
    stem(prob_history)
    title('Probability function value before the begin of any LBP iteration')
    xlabel('Iteration')
    ylabel('Probability')
    subplot(4,1,3)
    stem(data_prob_history)
    title('Data probability function value before the begin of any LBP iteration')
    xlabel('Iteration')
    ylabel('Data probability')
    subplot(4,1,4)
    stem(smooth_prob_history)
    title('Smoothness probability function value before the begin of any LBP iteration')
    xlabel('Iteration')
    ylabel('Smoothness probability')
    
end











