function [prob_value, data_prob_value, smooth_prob_value] = belief_propagation_probability_calculation(depthmap, interval, probability, neighborhood, TH, TEMPERATURE)

% Carlo Dal Mutto
%
% dalmutto@dei.unipd.it

global MIN_ROW_D
global MAX_ROW_D
global MIN_COL_D
global MAX_COL_D
global DEBUG 

data_prob = zeros(size(depthmap,1),size(depthmap,2)); % Data term, associated to the measured probability

prob_data_history = zeros(1,size(depthmap,1)*size(depthmap,2));
for i_r = 2:size(depthmap,1)-1 % Do not consider first and last row
    for i_c = 2:size(depthmap,2)-1 % Do not consider first and last column
        interval_temp = interval{i_r,i_c}(:);
        probability_temp = probability{i_r,i_c}(:);
        data_prob(i_r,i_c) = probability_temp(find(interval_temp==depthmap(i_r,i_c)));     
        prob_data_history((i_r-1)*size(depthmap,2)+i_c) = data_prob(i_r,i_c);
    end
end

smooth_prob = zeros(size(depthmap,1),size(depthmap,2)); % Smoothness term, associated to the smoothness cost function 
prob_smooth_history = zeros(1,size(depthmap,1)*size(depthmap,2));
for i_r = 2:size(depthmap,1)-1 % Do not consider first and last row
    for i_c = 2:size(depthmap,2)-1 % Do not consider first and last column
        
        smooth_cost_temp = 0;

        for i_n = 1: length(neighborhood{i_r,i_c})
            smooth_cost_temp = smooth_cost_temp + min([((depthmap(i_r,i_c)-depthmap( neighborhood{i_r,i_c}(i_n).v,neighborhood{i_r,i_c}(i_n).u))^2), TH^2]); 
        end
       
        smooth_prob(i_r,i_c) = -smooth_cost_temp/TEMPERATURE;
        prob_smooth_history((i_r-1)*size(depthmap,2)+i_c) = exp(-smooth_cost_temp/TEMPERATURE);
    end
end

probability_image = data_prob.*smooth_prob; % Image proportional to the per pixel probability of the measurements (ideally invertionally proportional to the error)

if DEBUG
    figure()
    subplot(2,2,1)
    imagesc(probability_image(MIN_ROW_D:MAX_ROW_D,MIN_COL_D:MAX_COL_D))%,[0 1])
    title('Probability Image')
    axis equal
end

% Calculating the probability value
probability_image = probability_image(2:size(probability_image,1)-1,2:size(probability_image,2)-1);
probability_vector = probability_image(:);
prob_value = mean(probability_vector);

% Calculating the data probability value
if DEBUG
    subplot(2,2,2)
    imagesc(data_prob(MIN_ROW_D:MAX_ROW_D,MIN_COL_D:MAX_COL_D),[0 1])
    title('Data Probability Image')
    axis equal
end
data_probability_image = data_prob(2:size(data_prob,1)-1,2:size(data_prob,2)-1);
data_probability_vector = data_probability_image(:);
data_prob_value = mean(data_probability_vector);

% Calculating the smoothness probability value
if DEBUG
    subplot(2,2,3)
    imagesc(smooth_prob(MIN_ROW_D:MAX_ROW_D,MIN_COL_D:MAX_COL_D))%,[0 1])
    title('Smoothness Probability Image')
    axis equal
end
smooth_probability_image = smooth_prob(2:size(smooth_prob,1)-1,2:size(smooth_prob,2)-1);
smooth_probability_vector = smooth_probability_image(:);
smooth_prob_value = mean(smooth_probability_vector);


if DEBUG
    subplot(2,2,4)
    imagesc(depthmap(MIN_ROW_D:MAX_ROW_D,MIN_COL_D:MAX_COL_D))
    title('depth map')
    axis equal
end

% cost_value = 0;
% for i = 1: length(prob_data_history)
%     cost_value = cost_value - log(prob_data_history(i)*prob_smooth_history(i));
% end


